@@ -1,29 +1,65 @@
-""" test doc string for AppBase.app_title tests. """
-
+""" unit and integration tests of the ae.core portion. """
 import datetime
 import logging
 import os
+import pytest
+import shutil
 import sys
 import threading
-from typing import cast, Any
 
-import pytest
 from conftest import delete_files
+from typing import cast, Any, TextIO
+from unittest.mock import patch
 
-from ae.base import DATE_TIME_ISO, force_encoding, read_file, write_file
+
+import ae.paths                 # for patch tests of ae.paths.PATH_PLACEHOLDERS
+import ae.core                  # for patch tests of ae.core.PATH_PLACEHOLDERS
+
+from ae.base import (
+    DATE_TIME_ISO,
+    force_encoding, norm_path, os_path_dirname, os_path_isfile, os_path_isdir, os_path_splitext, read_file, write_file)
+from ae.paths import PATH_PLACEHOLDERS, placeholder_path, coll_folders, Collector
 # noinspection PyProtectedMember
 from ae.core import (
-    APP_KEY_SEP, DEBUG_LEVELS, DEBUG_LEVEL_DISABLED, DEBUG_LEVEL_ENABLED, DEBUG_LEVEL_VERBOSE, MAX_NUM_LOG_FILES,
-    LOG_FILE_IDX_WIDTH,
-    activate_multi_threading, _deactivate_multi_threading, hide_dup_line_prefix, main_app_instance, po,
-    registered_app_names,
-    AppBase, _PrintingReplicator, SubApp)
+    APP_KEY_SEP, DEBUG_LEVELS, DEBUG_LEVEL_DISABLED, DEBUG_LEVEL_ENABLED, DEBUG_LEVEL_VERBOSE,
+    LOG_FILE_IDX_WIDTH, MAX_NUM_LOG_FILES,
+    activate_multi_threading, _deactivate_multi_threading, hide_dup_line_prefix, main_app_instance, print_out,
+    debug_out, verbose_out, is_debug, is_verbose, registered_app_names,
+    temp_context_cleanup, temp_context_folders, temp_context_get_or_create, _temp_folders,
+    AppBase, _PrintingReplicator)
 
 
 __version__ = '3.6.9dev-test'   # used for automatic app version find tests
 
 
 class TestCoreHelpers:
+    def test_debug_out(self, capsys, restore_app_env):
+        app = AppBase()
+        app.debug_level = DEBUG_LEVEL_VERBOSE
+
+        debug_out('tst console output')
+
+        out, err = capsys.readouterr()
+        assert 'tst console output' in out
+        assert err == ""
+
+        app.debug_level = DEBUG_LEVEL_ENABLED
+
+        debug_out('tst console output')
+
+        out, err = capsys.readouterr()
+        assert 'tst console output' in out
+        assert err == ""
+
+        app.debug_level = DEBUG_LEVEL_DISABLED
+
+        debug_out('tst console output')
+
+        out, err = capsys.readouterr()
+        assert 'tst console output' not in out
+        assert out == ""
+        assert err == ""
+
     def test_hide_dup_line_prefix(self):
         l1 = "<t_s_t>"
         l2 = l1
@@ -34,40 +70,68 @@ class TestCoreHelpers:
         l2 = l1[:3] + l1
         assert hide_dup_line_prefix(l1, l2) == " " * 3 + l1
 
+    def test_is_debug(self, restore_app_env):
+        app = AppBase()
+        app.debug_level = DEBUG_LEVEL_DISABLED
+
+        assert not is_debug()
+
+        app.debug_level = DEBUG_LEVEL_ENABLED
+
+        assert is_debug()
+
+        app.debug_level = DEBUG_LEVEL_VERBOSE
+
+        assert is_debug()
+
+    def test_is_verbose(self, restore_app_env):
+        app = AppBase()
+        app.debug_level = DEBUG_LEVEL_DISABLED
+
+        assert not is_verbose()
+
+        app.debug_level = DEBUG_LEVEL_ENABLED
+
+        assert not is_verbose()
+
+        app.debug_level = DEBUG_LEVEL_VERBOSE
+
+        assert is_verbose()
+
     def test_print_out_basics(self, capsys):
-        po()
+        print_out()
         out, err = capsys.readouterr()
         assert out == '\n' and err == ''
 
-        po(invalid_kwarg='ika')
+        print_out(invalid_kwarg='ika')
         out, err = capsys.readouterr()
         assert 'ika' in out and err == ''
 
         us = chr(40960) + chr(1972) + chr(2013) + 'äöü'
-        po(us, encode_errors_def='strict')
+        print_out(us, encode_errors_def='strict')
         out, err = capsys.readouterr()
         assert us in out and err == ''
 
-        po(us, file=sys.stdout)
-        po(us, file=sys.stderr)
+        print_out(us, file=sys.stdout)
+        print_out(us, file=sys.stderr)
         fna = 'print_out.txt'
         fhd = open(fna, 'w', encoding='ascii', errors='strict')
-        po(us, file=fhd)
+        print_out(us, file=fhd)
         fhd.close()
         assert delete_files(fna) == 1
-        po(bytes(chr(0xef) + chr(0xbb) + chr(0xbf), encoding='utf-8'))
+        print_out(bytes(chr(0xef) + chr(0xbb) + chr(0xbf), encoding='utf-8'))
         out, err = capsys.readouterr()
         assert us in out
         assert us in err
 
     def test_print_out_cov(self, capsys):
-        # print invalid/surrogate code point/char to force UnicodeEncodeError exception in po() (testing coverage)
+        # print invalid/surrogate code point/char to force UnicodeEncodeError exception in print_out() (test coverage)
         us = chr(0xD801)
-        po(us, 123456, encode_errors_def='strict')      # .. also coverage of not-str args
+        print_out(us, 123456, encode_errors_def='strict')      # .. also coverage of not-str args
         out, err = capsys.readouterr()
         assert force_encoding(us) in out and '123456' in out and err == ''
 
-        po('\r', 123456)     # coverage of processing output (not captured by pytest)
+        print_out('\r', 123456)     # coverage of processing output (not captured by pytest)
         out, err = capsys.readouterr()
         assert out == '' and err == ''
 
@@ -80,6 +144,34 @@ class TestCoreHelpers:
         assert len(registered_app_names()) == 1
         assert app.app_name == registered_app_names()[0]
 
+    def test_verbose_out(self, capsys, restore_app_env):
+        app = AppBase()
+        app.debug_level = DEBUG_LEVEL_VERBOSE
+
+        verbose_out('tst console output')
+
+        out, err = capsys.readouterr()
+        assert 'tst console output' in out
+        assert err == ""
+
+        app.debug_level = DEBUG_LEVEL_ENABLED
+
+        verbose_out('tst console output')
+
+        out, err = capsys.readouterr()
+        assert 'tst console output' not in out
+        assert out == ""
+        assert err == ""
+
+        app.debug_level = DEBUG_LEVEL_DISABLED
+
+        verbose_out('tst console output')
+
+        out, err = capsys.readouterr()
+        assert 'tst console output' not in out
+        assert out == ""
+        assert err == ""
+
 
 class TestPrintingReplicator:
     def test_init(self):
@@ -127,10 +219,8 @@ class TestPrintingReplicator:
             dso.write(msg)
             lfo.close()
             with open(lfn) as f:
-                if f.encoding == 'ascii':
-                    assert f.read() == '\\ua000\\u07b4'
-                else:
-                    assert f.read() == msg      # msg == '\ua000\u07b4'
+                assert f.encoding == 'UTF-8'
+                assert f.read() == msg              # == '\ua000\u07b4'
 
         finally:
             assert delete_files(lfn) == 1
@@ -146,7 +236,7 @@ class TestAeLogging:
             for idx in range(MAX_NUM_LOG_FILES + 9):
                 for line_no in range(16):                   # full loop is creating 1 kB of log entries (16 * 64 bytes)
                     app.po("TestBaseLogEntry{: >26}{: >26}".format(idx, line_no))
-            assert os.path.exists(log_file)
+            assert os_path_isfile(log_file)
         finally:
             assert delete_files(log_file, keep_ext=True) == MAX_NUM_LOG_FILES + 1
 
@@ -154,7 +244,7 @@ class TestAeLogging:
         log_file = 'test_ae_cov_log.log'
         valid_log_content = "TestBaseLogEntry"
         invalid_log_content = "NeverAppearInLogFile"
-        fb, ext = os.path.splitext(log_file)    # simulate left-over log file from last app run - coverage
+        fb, ext = os_path_splitext(log_file)    # simulate a left-over log file from the last app run - coverage
         idx = 1
         write_file(f"{fb}-{idx:0>{LOG_FILE_IDX_WIDTH}}{ext}",
                    f"log file content to test left-over from last app run{invalid_log_content}")
@@ -164,7 +254,7 @@ class TestAeLogging:
             for idx in range(MAX_NUM_LOG_FILES + 9):
                 for line_no in range(16):                   # full loop is creating 1 kB of log entries (16 * 64 bytes)
                     app.po(f"{valid_log_content}{idx: >26}{line_no: >26}")
-            assert os.path.exists(log_file)
+            assert os_path_isfile(log_file)
         finally:
             contents = delete_files(log_file, keep_ext=True, ret_type='contents')
             assert len(contents) == MAX_NUM_LOG_FILES + 1
@@ -184,7 +274,7 @@ class TestAeLogging:
             for idx in range(MAX_NUM_LOG_FILES + 9):
                 for line_no in range(16):
                     app.po("TestBaseLogEntry{: >26}{: >26}".format(idx, line_no))
-            assert os.path.exists(log_file)
+            assert os_path_isfile(log_file)
         finally:
             assert delete_files(log_file, keep_ext=True) == MAX_NUM_LOG_FILES + 1
 
@@ -198,7 +288,7 @@ class TestAeLogging:
             for idx in range(MAX_NUM_LOG_FILES + 9):
                 for line_no in range(16):
                     app.po("TestBaseLogEntry{: >26}{: >26}".format(idx, line_no))
-            assert os.path.exists(log_file)
+            assert os_path_isfile(log_file)
         finally:
             assert delete_files(log_file, keep_ext=True) == MAX_NUM_LOG_FILES + 1
 
@@ -212,8 +302,29 @@ class TestAeLogging:
             app.po(tst_out)
             out, err = capsys.readouterr()
             assert out == "" and err == ""
-            app.init_logging()      # close log file
-            assert os.path.exists(log_file)
+            app.init_logging()      # close the log file
+            assert os_path_isfile(log_file)
+            out, err = capsys.readouterr()
+            assert out == "" and err == ""
+        finally:
+            contents = delete_files(log_file, ret_type='contents')
+            assert len(contents)
+            assert tst_out in contents[0]
+
+    def test_open_log_file_with_suppressed_stdout_reopen(self, capsys, restore_app_env):
+        log_file = 'test_ae_no_stdout.log'
+        tst_out = 'only printed to log file'
+        try:
+            app = AppBase('test_open_log_file_with_suppressed_stdout', suppress_stdout=True)
+            app._nul_std_out.close()
+
+            assert app.suppress_stdout is True
+            app.init_logging(log_file_name=log_file)
+            app.po(tst_out)
+            out, err = capsys.readouterr()
+            assert out == "" and err == ""
+            app.init_logging()      # close the log file
+            assert os_path_isfile(log_file)
             out, err = capsys.readouterr()
             assert out == "" and err == ""
         finally:
@@ -227,7 +338,7 @@ class TestAeLogging:
         app.init_logging(log_file_name=log_file)
         with pytest.raises(FileNotFoundError):
             app.log_file_check()     # coverage of callee exception
-        assert not os.path.exists(log_file)
+        assert not os_path_isfile(log_file)
 
     def test_log_file_flush(self, restore_app_env):
         log_file = 'test_ae_base_log_flush.log'
@@ -235,7 +346,7 @@ class TestAeLogging:
             app = AppBase('test_base_log_file_flush')
             app.init_logging(log_file_name=log_file)
             app.log_file_check()
-            assert os.path.exists(log_file)
+            assert os_path_isfile(log_file)
         finally:
             assert delete_files(log_file) == 1
 
@@ -247,18 +358,18 @@ class TestAeLogging:
         try:
             app = AppBase('test_main_app')
             app.init_logging(log_file_name=mp + log_file)
-            sub = SubApp('test_sub_app', app_name=sp)
+            sub = AppBase('test_sub_app', app_name=sp)
             sub.init_logging(log_file_name=sp + log_file)
-            po(mp + tst_out + "_1")
+            print_out(mp + tst_out + "_1")
             app.po(mp + tst_out + "_2")
             sub.po(sp + tst_out)
             sub.init_logging()
-            app.init_logging()  # close log file
+            app.init_logging()  # close the log file
             # NOT WORKING: capsys.readouterr() returning empty strings
             # out, err = capsys.readouterr()
             # assert out.count(tst_out) == 3 and err == ""
-            assert os.path.exists(mp + log_file)
-            assert os.path.exists(sp + log_file)
+            assert os_path_isfile(mp + log_file)
+            assert os_path_isfile(sp + log_file)
         finally:
             contents = delete_files(sp + log_file, ret_type='contents')
             assert len(contents)
@@ -277,7 +388,7 @@ class TestAeLogging:
         def sub_app_po():
             """ test thread function """
             nonlocal sub, sub_printed
-            sub = SubApp('test_sub_app_thread', app_name=sp)
+            sub = AppBase('test_sub_app_thread', app_name=sp)
             sub.init_logging(log_file_name=sp + log_file)
             sub.po(sp + tst_out)
             sub_printed = True
@@ -293,14 +404,14 @@ class TestAeLogging:
             sub_thread = threading.Thread(target=sub_app_po)
             sub_thread.start()
             while not sub_printed:      # NOT ENOUGH fails on gitlab CI: not sub or not sub.active_log_stream:
-                pass                    # wait until sub-thread has called init_logging()
-            po(mp + tst_out + "_1")
+                pass                    # wait until the sub-thread has called init_logging()  # pragma: no cover
+            print_out(mp + tst_out + "_1")
             app.po(mp + tst_out + "_2")
-            sub.init_logging()          # close sub-app log file created by sub_thread
+            sub.init_logging()          # close the sub-app log file created by sub_thread
             sub_thread.join()
-            app.init_logging()          # close main-app log file
-            assert os.path.exists(sp + log_file)
-            assert os.path.exists(mp + log_file)
+            app.init_logging()          # close the main-app log file
+            assert os_path_isfile(sp + log_file)
+            assert os_path_isfile(mp + log_file)
         finally:
             contents = delete_files(sp + log_file, ret_type='contents')
             assert len(contents)
@@ -316,7 +427,8 @@ class TestAeLogging:
     def test_exception_log_file_flush(self, restore_app_env):
         app = AppBase('test_exception_base_log_file_flush')
         # cause/provoke _append_eof_and_flush_file() exceptions for coverage by passing any other non-stream object
-        app._append_eof_and_flush_file(cast('TextIO', None), 'invalid stream')
+        # noinspection PyInvalidCast
+        app._append_eof_and_flush_file(cast(TextIO, None), 'invalid stream')
 
     def test_app_instances_reset2(self):
         assert main_app_instance() is None
@@ -326,91 +438,95 @@ class TestPythonLogging:
     """ test python logging module support
     """
     def test_log_init(self, restore_app_env):
-        var_val = dict(version=1,
-                       disable_existing_loggers=False)
+        var_val = {'version': 1,
+                   'disable_existing_loggers': False}
         app = AppBase('log_init')
         app.init_logging(py_logging_params=var_val)
 
         assert app.py_log_params == var_val
 
-        logging.shutdown()
-
     def test_app_instances_reset1(self):
         assert main_app_instance() is None
 
     def test_logging_params_dict_console_from_init(self, restore_app_env):
-        var_val = dict(version=1,
-                       disable_existing_loggers=False,
-                       handlers=dict(console={'class': 'logging.StreamHandler',
-                                              'level': logging.INFO}))
+        var_val = {'version': 1,
+                   'disable_existing_loggers': False,
+                   'handlers': {'console': {'class': 'logging.StreamHandler',
+                                            'level': logging.INFO}}}
         print(str(var_val))
 
-        cae = AppBase('test_python_logging_params_dict_console')
+        main_app = AppBase('test_python_logging_params_dict_console')
+        main_app.init_logging(py_logging_params=var_val)
+
+        assert main_app.py_log_params == var_val
+
+    def test_logging_params_dict_console_from_init_with_log_file(self, restore_app_env):
+        log_file = 'tst_py_log_complex.log'
+        var_val = {'version': 1,
+                   'disable_existing_loggers': False,
+                   'handlers': {'console': {'class': 'logging.handlers.RotatingFileHandler',
+                                            'level': logging.INFO,
+                                            'filename': log_file,
+                                            'maxBytes': 33,
+                                            'backupCount': 63}},
+                   'loggers': {'root': {'handlers': ['console']},
+                               'ae': {'handlers': ['console']},
+                               'ae.console': {'handlers': ['console']}}}
+        print(str(var_val))
+
+        cae = AppBase('test_python_logging_params_dict_file', debug_level=DEBUG_LEVEL_DISABLED)
         cae.init_logging(py_logging_params=var_val)
 
-
         assert cae.py_log_params == var_val
-        logging.shutdown()
 
-    def test_logging_params_dict_complex(self, restore_app_env, sys_argv_app_key_restore):
-        """ test logging with rotating file handler, first refactored migrated from
+        # empty log file created by logging.config.dictConfig(py_logging_params) in :func:`ae.core.AppBase.init_logging`
+        assert delete_files(log_file, ret_type='contents')[0] == ""
 
-        TODO: investigate and fix the 4 commented out asserts in this test method
-        .. which found log_text via `grm check` pytest in console as well as in pycharm pytest run
-        .. but in this test module (test_core.py) only the grm/pytest run does not find log_text/log_files at all!
-        .. or shows them accumulated in a later test
-        Strange: a very similar test method did run fine with grm&pycharm in ae_console/tests/test_console.py (v0.3.63)
-
-        Looks like the problem lies in pytest (previous version done with caplog had also empty .text)
-        .. but also in PyCharm, see:
-
-        * https://stackoverflow.com/questions/59875983
-        * https://github.com/pytest-dev/pytest/issues/3697
-        * https://youtrack.jetbrains.com/issue/PY-48743/Running-Pytest-is-not-showing-logging-output
-
-
-        """
+    def test_logging_params_dict_complex(self, restore_app_env):
+        """ test logging with rotating file handler. """
         log_file = 'test_py_log_complex.log'
         entry_prefix = "TEST LOG ENTRY "
 
-        var_val = dict(version=1,
-                       disable_existing_loggers=False,
-                       handlers=dict(console={'class': 'logging.handlers.RotatingFileHandler',
-                                              'level': logging.INFO,
-                                              'filename': log_file,
-                                              'maxBytes': 33,
-                                              'backupCount': 63}),
-                       loggers={'root': dict(handlers=['console']),
-                                'ae': dict(handlers=['console']),
-                                'ae.console': dict(handlers=['console'])}
-                       )
+        var_val = {'version': 1,
+                   'disable_existing_loggers': False,
+                   'handlers': {'console': {'class': 'logging.handlers.RotatingFileHandler',
+                                            'level': logging.INFO,
+                                            'filename': log_file,
+                                            'maxBytes': 33,
+                                            'backupCount': 63}},
+                   'loggers': {'root': {'handlers': ['console']},
+                               'ae': {'handlers': ['console']},
+                               'ae.core': {'handlers': ['console']}}}
         print(str(var_val))
 
-        cae = AppBase('test_python_logging_params_dict_file')
-        cae.init_logging(py_logging_params=var_val)
-
+        cae = AppBase('test_python_logging_params_dict_file', debug_level=DEBUG_LEVEL_DISABLED)
+        cae.init_logging(py_logging_params=var_val)  # logging.config.dictConfig(py_logging_params) creates empty logFil
 
         assert cae.py_log_params == var_val
 
         root_logger = logging.getLogger()   # 'root'
         ae_logger = logging.getLogger('ae')
-        ae_cae_logger = logging.getLogger('ae.console')
+        ae_cae_logger = logging.getLogger('ae.core')
 
-        # ConsoleApp print_out
+        # AppBase print_out/.po
         try:
             log_text = entry_prefix + "0 print_out"
             cae.po(log_text)
         finally:
             logging.shutdown()
-            assert delete_files(log_file, ret_type='contents')[0] == ""
+            # empty log file created in log_init, and gets not extended by the above cae.po() call
+            files_contents = delete_files(log_file, ret_type='contents')
+            assert len(files_contents) == 1
+            assert files_contents[0] == ""
 
         try:
             log_text = entry_prefix + "0 print_out root"
             cae.po(log_text, logger=root_logger)
         finally:
             logging.shutdown()
-            # grm-pytest-run: assert delete_files(log_file) == 0
-            # pycharm-pytest-run: assert delete_files(log_file, ret_type='contents')[0].endswith(log_text + os.linesep)
+            files_contents = delete_files(log_file, ret_type='contents')
+            assert len(files_contents) == 1
+            assert files_contents[0].endswith(log_text + os.linesep)
 
         try:
             log_text = entry_prefix + "0 print_out ae"
@@ -426,14 +542,14 @@ class TestPythonLogging:
             cae.po(log_text, logger=ae_cae_logger)
         finally:
             logging.shutdown()
-            # multiple log files because log text has 34 bytes but RotatingFileHandler maxbytes is 33
+            # multiple log files because the log text has 34 bytes but RotatingFileHandler maxbytes is 33
             files_contents = delete_files(log_file, ret_type='contents')
             assert len(files_contents) > 1
             assert any(_.endswith(log_text + os.linesep) for _ in files_contents)
 
         # logging
         try:
-            logging.info(entry_prefix + "1 info")       # will NOT be added to log
+            logging.info(entry_prefix + "1 info")       # will NOT be added to the log
         finally:
             logging.shutdown()
             assert delete_files(log_file) == 0
@@ -449,16 +565,18 @@ class TestPythonLogging:
             logging.warning(log_text)
         finally:
             logging.shutdown()
-            # grm: assert delete_files(log_file) == 0
+            # pjm: assert delete_files(log_file) == 0
             # pycharm: assert delete_files(log_file, ret_type='contents')[0].endswith(log_text + os.linesep)
+            assert delete_files(log_file, ret_type='contents')[0].endswith(log_text + os.linesep)
 
         try:
             log_text = entry_prefix + "4 error logging"
             logging.error(log_text)
         finally:
             logging.shutdown()
-            # grm: assert delete_files(log_file) == 0
+            # pjm: assert delete_files(log_file) == 0
             # pycharm: assert delete_files(log_file, ret_type='contents')[0].endswith(log_text + os.linesep)
+            assert delete_files(log_file, ret_type='contents')[0].endswith(log_text + os.linesep)
 
         # loggers
         try:
@@ -466,16 +584,17 @@ class TestPythonLogging:
             root_logger.error(log_text)
         finally:
             logging.shutdown()
-            # grm: assert delete_files(log_file) == 0
+            # pjm: assert delete_files(log_file) == 0
             # pycharm: assert delete_files(log_file, ret_type='contents')[0].endswith(log_text + os.linesep)
+            assert delete_files(log_file, ret_type='contents')[0].endswith(log_text + os.linesep)
 
         try:
             log_text = entry_prefix + "4 error ae"
             ae_logger.error(log_text)
         finally:
             logging.shutdown()
-            # grm: assert delete_files(log_file, ret_type='contents')[0].endswith(log_text + os.linesep)
-            # grm+pycharm: delete_files returns 5 files (this one 2*, all the before missing ones & wrong ordered)?!?!?
+            # pjm: assert delete_files(log_file, ret_type='contents')[0].endswith(log_text + os.linesep)
+            # pjm+pycharm: delete_files returns 5 files (this one 2*, all the before missing ones & wrong ordered)?!?!?
             assert log_text + os.linesep in delete_files(log_file, ret_type='contents')
 
         try:
@@ -485,18 +604,19 @@ class TestPythonLogging:
             logging.shutdown()
             assert delete_files(log_file, ret_type='contents')[0].endswith(log_text + os.linesep)
 
-        # ConsoleAppEnv dpo
+        # AppBase.debug_out/.dpo
         sys.argv = ['tl_cdc']  # sys.argv has to be set to allow get_option('debug_level') calls done by debug_out()
         try:
-            log_text = entry_prefix + "5 not logged dpo"
-            cae.dpo(log_text, minimum_debug_level=DEBUG_LEVEL_DISABLED)
+            log_text = entry_prefix + "5 not logged debug_out/dpo"
+            cae.dpo(log_text)
         finally:
             logging.shutdown()
             assert delete_files(log_file) == 0
 
+        # AppBase.print_out/.po
         try:
-            log_text = entry_prefix + "5 dpo"
-            cae.dpo(log_text, minimum_debug_level=DEBUG_LEVEL_DISABLED, logger=ae_cae_logger)
+            log_text = entry_prefix + "5 print_out/po"
+            cae.po(log_text, logger=ae_cae_logger)
         finally:
             logging.shutdown()
             assert delete_files(log_file, ret_type='contents')[0].endswith(log_text + os.linesep)
@@ -505,16 +625,18 @@ class TestPythonLogging:
         assert main_app_instance() is None
 
 
-class TestAppBase:      # only some basic tests - test coverage is done by :class:`~.console.ConsoleApp` tests
-    def test_app_name(self, restore_app_env, sys_argv_app_key_restore):
+class TestAppBase:      # only some basic tests - test coverage is done by :class:`~ae.console.ConsoleApp` tests
+    def test_app_name(self, restore_app_env):
         name = 'tan_app_name'
         sys.argv = [name, ]
         app = AppBase()
         assert app.app_name == name
 
     def test_app_key(self, restore_app_env):
-        app = AppBase(app_name='XXX', sys_env_id='YYY')
-        assert app.app_key == 'XXX@YYY'
+        app_name = 'TstAppName'
+        env_id = 'TstEnvId'
+        app = AppBase(app_name=app_name, sys_env_id=env_id)
+        assert app.app_key == app_name + '@' + env_id
 
     def test_app_instances_reset1(self):
         assert main_app_instance() is None
@@ -525,7 +647,7 @@ class TestAppBase:      # only some basic tests - test coverage is done by :clas
         app = AppBase(title, app_version=ver)
         assert app.app_title == title
         assert app.app_version == ver
-        assert app.app_path == os.path.dirname(sys.argv[0])
+        assert app.app_path == norm_path(os_path_dirname(sys.argv[0]))
 
     def test_app_find_version(self, restore_app_env):
         app = AppBase()
@@ -550,7 +672,7 @@ class TestAppBase:      # only some basic tests - test coverage is done by :clas
 
         _instance = _OtherClass()
         t_args = (1, 2, '3')
-        t_kwargs = dict(a=1, b=2, c='3')
+        t_kwargs = {'a': 1, 'b': 2, 'c': '3'}
         assert app.call_method(_instance._method, *t_args, **t_kwargs) == (_instance, t_args, t_kwargs)
 
     def test_call_method_no_exception_if_not_exists(self, restore_app_env):
@@ -594,6 +716,97 @@ class TestAppBase:      # only some basic tests - test coverage is done by :clas
         setattr(app, 'test_raiser', _raising_ex)
         app.call_method('test_raiser')
 
+    def test_debug_out(self, capsys, restore_app_env):
+        app = AppBase(debug_level=DEBUG_LEVEL_ENABLED)
+        tst = "tsT-debug-out-string"
+        app.debug_out(tst)
+        assert tst in capsys.readouterr()[0]
+
+    def test_init_path_placeholders(self, restore_app_env):
+        cae = AppBase("test_init_path_placeholders")
+
+        assert PATH_PLACEHOLDERS is ae.core.PATH_PLACEHOLDERS
+        assert PATH_PLACEHOLDERS is ae.paths.PATH_PLACEHOLDERS
+
+        tst_path_placeholders = ae.paths.PATH_PLACEHOLDERS.copy()
+        with patch('ae.core.PATH_PLACEHOLDERS', tst_path_placeholders):
+            assert cae.app_name == 'pyTstConsAppKey'
+            assert ae.core.PATH_PLACEHOLDERS['app_name'] == 'pyTstConsAppKey'
+            assert ae.core.PATH_PLACEHOLDERS['main_app_name'] == 'pyTstConsAppKey'
+
+            new_app_name = 'some_new_app_name'
+            cae.app_name = new_app_name
+            cae._init_path_placeholders()
+            assert ae.core.PATH_PLACEHOLDERS['app_name'] == new_app_name
+            assert ae.core.PATH_PLACEHOLDERS['main_app_name'] == new_app_name
+            cae.app_name = 'pyTstConsAppKey'
+
+    def test_init_path_placeholders_read_err(self, restore_app_env):
+        cae = AppBase("test_init_path_placeholders_read_err")
+        usr_path = ae.core.PATH_PLACEHOLDERS['usr']
+        old_phs = ae.core.PATH_PLACEHOLDERS.copy()
+
+        with patch('ae.core.read_file', lambda *_args, **_kwargs: ""):
+            cae._init_path_placeholders()
+        assert ae.core.PATH_PLACEHOLDERS == old_phs
+        assert ae.paths.PATH_PLACEHOLDERS == old_phs
+
+        def _raise_err(file_path: str, *_args, **_kwargs):
+            if placeholder_path(file_path).startswith('{usr}'):
+                return read_file(file_path, *_args, **_kwargs)
+            else:
+                raise Exception("TstInitPathPlaceHoldersWriteErr")
+
+        with (patch('ae.core.read_file', _raise_err)):
+            tst_path_placeholders = ae.core.PATH_PLACEHOLDERS.copy()
+            with patch('ae.core.PATH_PLACEHOLDERS', tst_path_placeholders):
+                assert not ae.core.PATH_PLACEHOLDERS['ado'].startswith(usr_path)
+                assert not placeholder_path(ae.core.PATH_PLACEHOLDERS['ado']).startswith("{usr}")
+                try:
+                    cae._init_path_placeholders()
+                finally:
+                    for re_dir in Collector(item_collector=coll_folders
+                                            ).collect(usr_path, select=cae.app_name + "*").paths:
+                        shutil.rmtree(re_dir)
+                assert ae.core.PATH_PLACEHOLDERS['ado'].startswith(usr_path)
+                assert placeholder_path(ae.core.PATH_PLACEHOLDERS['ado']).startswith("{usr}")
+                assert ae.core.PATH_PLACEHOLDERS != old_phs
+
+            assert ae.core.PATH_PLACEHOLDERS == old_phs
+
+    def test_init_path_placeholders_write_err(self, restore_app_env):
+        cae = AppBase("test_init_path_placeholders_write_err")
+        usr_path = ae.core.PATH_PLACEHOLDERS['usr']
+        old_phs = ae.core.PATH_PLACEHOLDERS.copy()
+
+        with patch('ae.core.write_file', lambda *_args, **_kwargs: None):
+            cae._init_path_placeholders()
+        assert ae.core.PATH_PLACEHOLDERS == old_phs
+        assert ae.paths.PATH_PLACEHOLDERS == old_phs
+
+        def _raise_write_err(file_path: str, *_args, **_kwargs):
+            if placeholder_path(file_path).startswith('{usr}'):
+                write_file(file_path, *_args, **_kwargs)
+            else:
+                raise Exception("TstInitPathPlaceHoldersWriteErr")
+
+        with patch('ae.core.write_file', _raise_write_err):
+            tst_path_placeholders = ae.core.PATH_PLACEHOLDERS.copy()
+            with patch('ae.core.PATH_PLACEHOLDERS', tst_path_placeholders):
+                assert not ae.core.PATH_PLACEHOLDERS['ado'].startswith(usr_path)
+                assert not placeholder_path(ae.core.PATH_PLACEHOLDERS['ado']).startswith("{usr}")
+                try:
+                    cae._init_path_placeholders()
+                finally:
+                    for re_dir in Collector(item_collector=coll_folders
+                                            ).collect(usr_path, select=cae.app_name + "*").paths:
+                        shutil.rmtree(re_dir)
+                assert ae.core.PATH_PLACEHOLDERS['ado'].startswith(usr_path)
+                assert placeholder_path(ae.core.PATH_PLACEHOLDERS['ado']).startswith("{usr}")
+                assert ae.core.PATH_PLACEHOLDERS != old_phs
+
+            assert ae.core.PATH_PLACEHOLDERS == old_phs
+
     def test_log_line_prefix(self, restore_app_env):
         app = AppBase(sys_env_id='Tee sst')
         app._log_with_timestamp = True
@@ -632,26 +845,74 @@ class TestAppBase:      # only some basic tests - test coverage is done by :clas
         out, err = capsys.readouterr()
         assert us in out and us in err
 
-        # print invalid/surrogate code point/char to force UnicodeEncodeError exception in po() (testing coverage)
+        # print invalid/surrogate code point/char to force UnicodeEncodeError exception in print_out() (test coverage)
         us = chr(0xD801)
         app.po(us, encode_errors_def='strict')
 
-        # multi_threading has to be reset to prevent debug test run freeze (added multi_threading for coverage)
+        # multi_threading has to be reset to prevent a debug test run freeze (added multi_threading for coverage)
         _deactivate_multi_threading()
 
     def test_app_instances_reset2(self):
         assert main_app_instance() is None
 
+    def test_shutdown_err_msg(self, capsys, restore_app_env):
+        app = AppBase()
+
+        app.shutdown(None, error_message='AppBase.shutdown() error-message')
+
+        out, err = capsys.readouterr()
+        assert 'AppBase.shutdown() error-message' in out
+        assert err == ""
+
+    def test_shutdown_none(self, capsys, restore_app_env):
+        app = AppBase(debug_level=DEBUG_LEVEL_DISABLED)
+
+        app.shutdown(None)
+
+        out, err = capsys.readouterr()
+        assert out == ""
+        assert err == ""
+
+    def test_shutdown_0(self, capsys, restore_app_env):
+        app = AppBase(debug_level=DEBUG_LEVEL_ENABLED)
+
+        with patch('ae.core.sys.exit', lambda *args, **kwargs: None):
+            app.shutdown()  # exit_code == 0
+
+        out, err = capsys.readouterr()
+        assert 'shutdown of ' in out
+        assert err == ""
+
+    def test_shutdown_1(self, capsys, restore_app_env):
+        app = AppBase(debug_level=DEBUG_LEVEL_VERBOSE)
+
+        with patch('ae.core.sys.exit', lambda *args, **kwargs: None):
+            app.shutdown(1)
+
+        assert 'shutdown of ' in capsys.readouterr()[0]
+
+    def test_shutdown_123456(self, capsys, restore_app_env):
+        app = AppBase()
+
+        with patch('ae.core.sys.exit', lambda *args, **kwargs: None):
+            app.shutdown(123456)    # test warning if exit code is not in 0..255
+
+        assert 'extended exit code' in capsys.readouterr()[0]
+
+    def test_shutdown_temp_context(self, restore_app_env):
+        app = AppBase(debug_level=DEBUG_LEVEL_ENABLED)
+        assert not _temp_folders
+        temp_context_get_or_create('tst_shutdown_tmp_context')
+        assert 'tst_shutdown_tmp_context' in _temp_folders
+
+        app.shutdown(exit_code=None)
+
+        assert not _temp_folders
+
     def test_verbose(self, capsys, restore_app_env):
         app = AppBase(debug_level=DEBUG_LEVEL_VERBOSE)
         assert app.verbose
 
-    def test_debug_out(self, capsys, restore_app_env):
-        app = AppBase(debug_level=DEBUG_LEVEL_ENABLED)
-        tst = "tsT-debug-out-string"
-        app.debug_out(tst)
-        assert tst in capsys.readouterr()[0]
-
     def test_verbose_out(self, capsys, restore_app_env):
         app = AppBase(debug_level=DEBUG_LEVEL_ENABLED)
         tst = "tsT-debug-miss-string"
@@ -662,3 +923,76 @@ class TestAppBase:      # only some basic tests - test coverage is done by :clas
         tst = "tsT-verbose-out-string"
         app.verbose_out(tst)
         assert tst in capsys.readouterr()[0]
+
+
+class TestTempContextDirectories:
+    def test_temp_context_cleanup(self):
+        path = temp_context_get_or_create()
+        assert path == temp_context_get_or_create()
+        assert os_path_isdir(path)
+
+        temp_context_cleanup()
+
+        assert not os_path_isdir(path)
+
+        new_path = temp_context_get_or_create()
+        assert os_path_isdir(new_path)
+        assert new_path != path
+
+        temp_context_cleanup()
+
+        assert not os_path_isdir(new_path)
+
+    def test_temp_context_folders(self):
+        folder_name = "tst_tmp_dir"
+        path = temp_context_get_or_create(folder_name=folder_name)
+
+        assert path.endswith(folder_name)
+        assert set(temp_context_folders()) == {folder_name}
+
+        assert temp_context_folders(context="any not existing context") == []
+
+    def test_temp_context_get_or_create(self):
+        path = temp_context_get_or_create()
+        assert os_path_isdir(path)
+        temp_context_cleanup()
+
+    def test_temp_context_get_or_create_named(self):
+        ctx_name = "any string to name a temp dir"
+
+        path = temp_context_get_or_create(context=ctx_name)
+
+        assert os_path_isdir(path)
+        temp_context_cleanup()
+        assert os_path_isdir(path)
+        temp_context_cleanup(context=ctx_name)
+        assert not os_path_isdir(path)
+
+    def test_temp_context_get_or_create_with_folder(self):
+        dir1 = "name of the first temp dir"
+        dir2 = "TempDirFolder2"
+
+        path = temp_context_get_or_create(folder_name=dir1)
+
+        assert path.endswith(dir1)
+        assert os_path_isdir(path)
+
+        path2 = temp_context_get_or_create(folder_name=dir2)
+
+        assert path2.endswith(dir2)
+        assert os_path_isdir(path)
+        assert os_path_isdir(path2)
+        assert set(temp_context_folders()) == {dir1, dir2}
+
+        path2a = temp_context_get_or_create(folder_name=dir2)
+
+        assert path2a == path2
+        assert path2a.endswith(dir2)
+        assert os_path_isdir(path)
+        assert os_path_isdir(path2a)
+        assert set(temp_context_folders()) == {dir1, dir2}
+
+        temp_context_cleanup()
+
+        assert not os_path_isdir(path)
+        assert not os_path_isdir(path2)
